"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var Debugger = require("debug");
var debug = Debugger('akita:json-stream');
function uint8ArrayToString(data) {
    if (typeof data === 'string')
        return data;
    if (typeof Buffer !== 'undefined' && Buffer.isBuffer(data))
        return data.toString('utf-8');
    if (typeof TextDecoder !== 'undefined')
        return new TextDecoder('utf-8').decode(data);
    var str = '';
    for (var i = 0; i < data.length; i += 1) {
        str += String.fromCharCode(data[i]);
    }
    return str;
}
var JsonStream = (function () {
    function JsonStream(stream, reducer) {
        var _this = this;
        this._receive = function (data) {
            var string = uint8ArrayToString(data);
            debug('receive', string);
            if (_this.closed)
                return;
            _this._cache += string;
            _this._parse();
        };
        this._onError = function (error) {
            if (_this.listenerCount('error')) {
                _this.emit('error', error);
            }
            if (_this._reject) {
                _this._reject(error);
                _this._resolve = null;
                _this._reject = null;
            }
            if (!_this.closed)
                _this._close();
        };
        this._close = function () {
            debug('_close');
            if (_this.closed)
                return;
            _this.closed = true;
            if (_this.listenerCount('close')) {
                _this.emit('close');
            }
            if (_this._resolve) {
                _this._resolve(undefined);
                _this._resolve = null;
                _this._reject = null;
            }
            if (!_this._stream.getReader) {
                _this._stream.removeListener('data', _this._receive);
                _this._stream.removeListener('close', _this._close);
            }
            delete _this._stream;
            delete _this._reader;
            delete _this._cache;
            delete _this._reducer;
            _this._listeners = {};
        };
        this.closed = false;
        this._stream = stream;
        this._queue = [];
        this._cache = '';
        this._reducer = reducer;
        this._listeners = {};
        if (stream.getReader) {
            this._reader = stream.getReader();
            var read_1 = function (_a) {
                var done = _a.done, value = _a.value;
                if (_this.closed)
                    return;
                _this._receive(value || (done && _this._cache ? '\n' : ''));
                if (done) {
                    _this._close();
                }
                else {
                    _this._reader.read().then(read_1, _this._onError);
                }
            };
            this._reader.read().then(read_1, this._onError);
        }
        else {
            stream.on('data', this._receive);
            stream.on('close', this._close);
        }
    }
    JsonStream.prototype._parse = function () {
        if (this.closed || !this._cache)
            return;
        var index = this._cache.indexOf('\n');
        if (index < 0)
            return;
        var line = this._cache.substr(0, index).trim();
        this._cache = this._cache.substr(index + 1);
        if (line) {
            var json = void 0;
            try {
                json = JSON.parse(line);
            }
            catch (e) {
                this._onError(e);
                return;
            }
            if (this._reducer && json.object) {
                json.object = this._reducer(json.object);
            }
            if (this.listenerCount('data')) {
                while (this._queue.length) {
                    this.emit('data', this._queue.shift());
                }
                this.emit('data', json);
            }
            if (this._resolve) {
                this._resolve(json);
                this._resolve = null;
                this._reject = null;
            }
            else if (!this.listenerCount('data')) {
                this._queue.push(json);
            }
            this._parse();
        }
    };
    JsonStream.prototype.on = function (name, fn) {
        if (!this._listeners[name]) {
            this._listeners[name] = [];
        }
        this._listeners[name].push(fn);
        return this;
    };
    JsonStream.prototype.emit = function (name) {
        var args = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            args[_i - 1] = arguments[_i];
        }
        if (!this._listeners[name])
            return;
        this._listeners[name].forEach(function (fn) {
            fn.apply(null, args);
        });
    };
    JsonStream.prototype.listenerCount = function (name) {
        return (this._listeners[name] || []).length;
    };
    JsonStream.prototype.removeListener = function (name, fn) {
        if (!this._listeners[name])
            return;
        this._listeners[name] = this._listeners[name].filter(function (f) { return f !== fn; });
        return this;
    };
    JsonStream.prototype.removeAllListeners = function (name) {
        delete this._listeners[name];
        return this;
    };
    JsonStream.prototype.read = function () {
        var _this = this;
        if (this._queue.length) {
            return Promise.resolve(this._queue.shift());
        }
        if (this.closed)
            return Promise.reject(new Error('Can not read from closed stream.'));
        return new Promise(function (resolve, reject) {
            _this._resolve = resolve;
            _this._reject = reject;
        });
    };
    JsonStream.prototype.cancel = function () {
        if (this.closed)
            throw new Error('Can not cancel closed stream.');
        debug('cancel watch');
        if (this._stream.cancel) {
            this._stream.cancel();
        }
        else {
            this._stream.destroy();
            this._stream.removeListener('data', this._receive);
        }
        this._close();
    };
    return JsonStream;
}());
exports.default = JsonStream;
