"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var Debugger = require("debug");
var qs = require("qs");
var json_stream_1 = require("./json-stream");
var utils_1 = require("./utils");
var debug = Debugger('akita:request');
function execHooks(request, hooks) {
    var promise;
    if (Array.isArray(hooks)) {
        hooks.forEach(function (h) {
            if (promise) {
                promise = promise.then(function () {
                    var p = h(request);
                    return p ? p : Promise.resolve();
                });
            }
            else {
                promise = h(request);
            }
        });
    }
    else {
        promise = hooks(request);
    }
    if (promise) {
        return promise.then(function () { return Promise.resolve(); });
    }
}
var Request = (function () {
    function Request(client, fetch, url, init, query, reducer) {
        var _this = this;
        this.client = client;
        this._query = query;
        this._reducer = reducer;
        this._fetch = fetch;
        this.url = url;
        this.init = init || {};
        this._steps = 0;
        var promise;
        if (init.body && client._options.onEncode) {
            debug('exec onEncode');
            promise = execHooks(this, client._options.onEncode);
        }
        if (client._options.onRequest) {
            debug('exec onRequest');
            if (promise) {
                promise = promise.then(function () { return execHooks(_this, client._options.onRequest); });
            }
            else {
                promise = execHooks(this, client._options.onRequest);
            }
        }
        if (promise) {
            promise = promise.then(function () { return _this._send(); });
        }
        else {
            promise = this._send();
        }
        this._responsePromise = promise;
    }
    Request.prototype._addStep = function () {
        this._steps += 1;
        if (this._steps >= 3) {
            this._endAt = Date.now();
        }
        this.client._updateProgress();
    };
    Request.prototype._end = function () {
        this._endAt = Date.now();
        this.client._updateProgress();
    };
    Request.prototype._send = function () {
        var _this = this;
        var _a = this, client = _a.client, url = _a.url, init = _a.init;
        var headers = init.headers || {};
        if (client._options.init) {
            init = Object.assign({}, client._options.init, init);
            if (client._options.init.headers) {
                headers = Object.assign({}, client._options.init.headers, headers);
            }
        }
        init.headers = headers;
        this.init = init;
        if (debug.enabled) {
            debug(init.method, url, JSON.stringify(init));
        }
        if (init.body && typeof init.body === 'object') {
            init.body = client.createBody(init.body);
            var FormData_1 = client.getFormDataClass();
            if (!utils_1.isUint8Array(init.body) && !(FormData_1 && init.body instanceof FormData_1)) {
                if (!init.headers) {
                    init.headers = {};
                }
                if (!init.headers['Content-Type'] || init.headers['Content-Type'].indexOf('json') > -1) {
                    if (!init.headers['Content-Type']) {
                        init.headers['Content-Type'] = 'application/json';
                    }
                    init.body = JSON.stringify(init.body);
                }
                else if (init.headers['Content-Type'] === 'application/x-www-form-urlencoded') {
                    init.body = qs.stringify(init.body);
                }
                else {
                    throw new Error("Akita Error: Unsupported Content-Type " + init.headers['Content-Type']);
                }
            }
        }
        this._addStep();
        if (debug.enabled) {
            debug('fetch', url, JSON.stringify(init));
        }
        var promise = this._fetch(url, init).then(function (res) {
            _this._addStep();
            debug('response:', _this.url, res.status, res.statusText);
            _this.res = res;
            if (client._options.onResponse) {
                debug('exec onResponse');
                var p = execHooks(_this, client._options.onResponse);
                if (p) {
                    return p.then(function () {
                        return Promise.resolve(res);
                    });
                }
            }
            return Promise.resolve(res);
        }, function (error) {
            debug('fetch error:', error.message);
            _this._end();
            return Promise.reject(error);
        });
        return promise;
    };
    Request.prototype.response = function () {
        return this._responsePromise;
    };
    Request.prototype.stream = function () {
        var _this = this;
        if (this.res)
            return Promise.resolve(this.res.body);
        return this._responsePromise.then(function (res) {
            _this._end();
            return res.body;
        });
    };
    Request.prototype.jsonStream = function () {
        var _this = this;
        if (!this._jsPromise) {
            this._jsPromise = this.stream().then(function (stream) { return new json_stream_1.default(stream, _this._reducer); });
        }
        return this._jsPromise;
    };
    Request.prototype.ok = function () {
        if (this.res)
            return Promise.resolve(this.res.ok);
        return this._responsePromise.then(function (res) { return res.ok; });
    };
    Request.prototype.status = function () {
        if (this.res)
            return Promise.resolve(this.res.status);
        return this._responsePromise.then(function (res) { return res.status; });
    };
    Request.prototype.statusText = function () {
        if (this.res)
            return Promise.resolve(this.res.statusText);
        return this._responsePromise.then(function (res) { return res.statusText; });
    };
    Request.prototype.size = function () {
        if (this.res)
            return Promise.resolve(this.res.size || parseInt(this.res.headers.get('Content-Length')) || 0);
        return this._responsePromise.then(function (res) { return res.size || parseInt(res.headers.get('Content-Length')) || 0; });
    };
    Request.prototype.headers = function () {
        if (this.res)
            return Promise.resolve(this.res.headers);
        return this._responsePromise.then(function (res) { return res.headers; });
    };
    Request.prototype.buffer = function () {
        var _this = this;
        if (!this._bufferPromise) {
            this._bufferPromise = this._responsePromise.then(function (res) {
                var fn = 'buffer';
                if (res.arrayBuffer) {
                    fn = 'arrayBuffer';
                }
                return res[fn]().then(function (buf) {
                    if (!utils_1.isUint8Array(buf)) {
                        if (typeof Buffer === 'function') {
                            return Buffer.from(buf);
                        }
                        return buf;
                    }
                    _this._addStep();
                    return buf;
                }, function (error) {
                    debug('get buffer error:', error.message);
                    _this._end();
                    return Promise.reject(error);
                });
            });
        }
        return this._bufferPromise;
    };
    Request.prototype.blob = function () {
        var _this = this;
        if (!this._blobPromise) {
            this._blobPromise = this._responsePromise.then(function (res) {
                return res.blob().then(function (blob) {
                    _this._addStep();
                    return blob;
                }, function (error) {
                    debug('get blob error:', error.message);
                    _this._end();
                    return Promise.reject(error);
                });
            });
        }
        return this._blobPromise;
    };
    Request.prototype.text = function () {
        var _this = this;
        if (!this._textPromise) {
            if (this.raw) {
                this._textPromise = Promise.resolve(this.raw);
            }
            else {
                this._textPromise = this._responsePromise.then(function (res) {
                    return res.text().then(function (text) {
                        debug('response text:', text);
                        _this.raw = text;
                        _this._addStep();
                        return text;
                    }, function (error) {
                        debug('get text error:', error.message);
                        _this._end();
                        return Promise.reject(error);
                    });
                });
            }
        }
        return this._textPromise;
    };
    Request.prototype._decode = function () {
        if (this.value !== undefined)
            return this.value;
        debug('_decode');
        var value;
        try {
            value = JSON.parse(this.raw);
        }
        catch (error) {
            return Promise.reject(new Error("invalid json response body at " + this.url + " " + error.message));
        }
        if ((value === null || value === void 0 ? void 0 : value.error) && ['0', 'null', 'none'].indexOf(value.error) < 0) {
            var error_1 = new Error(value.error);
            Object.keys(value).forEach(function (key) {
                if (['error', 'message', 'stack'].indexOf(key) === -1) {
                    error_1[key] = value[key];
                }
            });
            return Promise.reject(error_1);
        }
        this.value = value;
        return value;
    };
    Request.prototype.data = function () {
        var _this = this;
        if (!this._dataPromise) {
            if (this.value !== undefined) {
                this._dataPromise = Promise.resolve(this.value);
            }
            else {
                this._dataPromise = this._responsePromise.then(function (res) {
                    if (res.status === 204) {
                        _this._addStep();
                        return Promise.resolve(null);
                    }
                    return res.text().then(function (text) {
                        _this._end();
                        debug('response text:', text);
                        if (res.status === 404 && _this._query && ['findByPk', 'remove'].indexOf(_this._query._op) > -1) {
                            debug("return null when " + _this._query._op + " 404");
                            return null;
                        }
                        _this.raw = text;
                        var client = _this.client;
                        if (client._options.onDecode) {
                            debug('exec onDecode');
                            var promise = execHooks(_this, client._options.onDecode);
                            if (promise) {
                                return promise.then(function () { return Promise.resolve(_this._decode()); });
                            }
                        }
                        return _this._decode();
                    }, function (error) {
                        debug('get text raw error:', error.message);
                        _this._end();
                        return Promise.reject(error);
                    });
                });
            }
        }
        return this._dataPromise;
    };
    Request.prototype.then = function (onSuccess, onFail) {
        var _this = this;
        if (this._query && this._query._op === 'watch') {
            return this.jsonStream().then(onSuccess, onFail);
        }
        if (this._reducer) {
            return this.data().then(function (json) { return onSuccess(_this._reducer(json)); }, onFail);
        }
        return this.data().then(onSuccess, onFail);
    };
    Request.prototype.catch = function (onFail) {
        return this.data().catch(onFail);
    };
    Request.prototype.finally = function (fn) {
        return this.data().finally(fn);
    };
    return Request;
}());
exports.default = Request;
